// Copyright 2018 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

//go:build sw64
// +build sw64

#include "go_asm.h"
#include "textflag.h"

TEXT ·Count(SB),NOSPLIT,$0-40
	LDL	R0, b_base+0(FP)
	LDL	R2, b_len+8(FP)
	LDBU	R1, c+24(FP)
	LDI	R8, $ret+32(FP)
	BR	countbytebody<>(SB)

TEXT ·CountString(SB),NOSPLIT,$0-32
	LDL	R0, s_base+0(FP)
	LDL	R2, s_len+8(FP)
	LDBU	R1, c+16(FP)
	LDI	R8, $ret+24(FP)
	BR	countbytebody<>(SB)

// input:
//   R0: data
//   R2: data len
//   R1: byte to find
//   R8: address to put result
TEXT countbytebody<>(SB),NOSPLIT,$0-0
	// R11 = count of byte to search
	LDI	R11, $0
	// short path to handle 0-byte case
	BEQ	R2, done
	CMPULT	R2, $0x10, R28
	// jump directly to tail if length < 32
	BNE	R28, tail

	AND	R0, $0xf, R9
	BEQ	R9, chunk
	// Work with not 32-byte aligned head
	BIC	R0, $0xf, R3
	ADDL	R3, $0x10, R3
head_loop:
	LDBU	R5, (R0)
	ADDL	R0, $1, R0
	CMPEQ	R5, R1, R28
	ADDL	R11, R28, R11
	SUBL	R2, $1, R2
	CMPEQ	R0, R3, R28
	BEQ	R28, head_loop
	// Work with 32-byte aligned chunks
chunk:
	BIC	R2, $0xf, R9
	// The first chunk can also be the last
	BEQ	R9, tail
	// R3 = end of 32-byte chunks
	ADDL	R0, R9, R3
	// R2 = length of tail
	SUBL	R2, R9, R2
	// Duplicate R1 (byte to search) to 16 1-byte elements of V0
	XOR R7, R7, R7
	BIS R1, R7, R7
	SLL R7, $8, R7
	BIS R1, R7, R7
	SLL R7, $8, R7
	BIS R1, R7, R7
	SLL R7, $8, R7
	BIS R1, R7, R7
	SLL R7, $8, R7
	BIS R1, R7, R7
	SLL R7, $8, R7
	BIS R1, R7, R7
	SLL R7, $8, R7
	BIS R1, R7, R7
	SLL R7, $8, R7
	BIS R1, R7, R7

	XOR R22, R22, R22
	// Count the target byte in 32-byte chunk
chunk_loop:
	LDL R16, (R0)
	LDL R17, 8(R0)
  ADDL R0, $16, R0
	CMPEQ	R0, R3, R28
	//VCMPUEQB	V0, V1, V3
	CMPGEB R7, R16, R18
	CMPGEB R7, R17, R19
	CMPGEB R16, R7, R16
	CMPGEB R17, R7, R17
	AND R16, R18, R16
	AND R17, R19, R17
	// Count lanes match the requested byte
	CTPOP	R16, R16 // 32B->16B
	CTPOP R17, R17
	ADDL	R16, R17, R16
	ADDL	R16, R22, R22
	// Accumulate the count in low 64-bit element of V8 when inside the loop
	BEQ	R28, chunk_loop
	ADDL	R22, R11, R11
	BEQ	R2, done
tail:
	// Work with tail shorter than 32 bytes
	LDBU	R5, (R0)
	ADDL R0, $1, R0
	SUBL R2, $1, R2
	CMPEQ	R5, R1, R28
	ADDL	R11, R28, R11
	BNE	R2, tail
done:
	STL	R11, (R8)
	RET
