// Copyright 2018 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

//go:build ignore
// +build ignore

#include "go_asm.h"
#include "textflag.h"

TEXT ·IndexByte(SB),NOSPLIT,$0-40
	LDL	R0, b_base+0(FP)
	LDL	R2, b_len+8(FP)
	LDBU	R1, c+24(FP)
	LDI	R8, $ret+32(FP)
	BR	indexbytebody<>(SB)

TEXT ·IndexByteString(SB),NOSPLIT,$0-32
	LDL R0,	s_base+0(FP)
	LDL R2,	s_len+8(FP)
	LDBU	R1, c+16(FP)
	LDI	R8, $ret+24(FP)
	BR	indexbytebody<>(SB)

// input:
//   R0: data
//   R1: byte to search
//   R2: data len
//   R8: address to put result
TEXT indexbytebody<>(SB),NOSPLIT,$0
	BEQ R2, notfound
	IFMOVD R1, F1
	// short path to handle 0-byte case
	CMPULT	R2, $0x20, R28
	BIS R0, R0, R16
	ADDL R2, R0, R3
	// jump directly to tail if length < 32
	// mov orignal data ptr to R16
	// Work with 32-byte aligned chunks
	VCPYB F1, V7

	BNE	R28, small
	// select page addr
	ADDL R2, R0, R3
	SUBL R3, $0x20, R3

chunk_loop:
	VLDD V5, (R0)
	VCMPUEQB V5, V7, V3
	// Count lanes match the requested byte
	FBNE F3, found
	// Accumulate the count in low 64-bit element of V8 when inside the loop
  ADDL R0, $0x20, R0
	CMPLE	R0, R3, R28
	BNE	R28, chunk_loop

	LDI R0, R3
	VLDD V5, (R3)
	VCMPUEQB V5, V7, V3
	// Count lanes match the requested byte
	FBNE F3, found

notfound:
	LDI R1, -1
	STL R1, (R8)
	RET

small:
	LDI R9, $8160
	ADDL R0, $0x20, R6
	AND R6, R9, R28
	BEQ R28, endofpage
	// Work with tail shorter than 32 bytes
	VLDD	V5, (R0)
	VCMPUEQB V5, V7, V3
	FBEQ F3, notfound
	FIMOVD F3, R5
	CTTZ R5, R5
	CMPLT R5, R2, R28
	BEQ R28, notfound
	STL R5, (R8)
	RET

found:
	//SUBL R0, $8, R0
	FIMOVD F3, R5
	CTTZ R5, R5
	SUBL R0, R16, R0
	ADDL R5, R0, R0
	CMPLT R0, R2, R28
	BEQ R28, notfound
	STL	R0, (R8)
	RET

endofpage:
	ADDL R0, R2, R3
	SUBL R3, $0x20, R3
	VLDD V5, (R3)
	VCMPUEQB V5, V7, V3
	FIMOVD F3, R5
	ZAPNOT R2, $1, R9
	SLL R5, R9, R5
	SRL R5, $0x20, R5
	BEQ R5, notfound
	CTTZ R5, R5
	STL R5, (R8)
	RET


